#!/usr/bin/env python3
import os
import json
import subprocess
import shutil
from datetime import datetime
from argparse import ArgumentParser
from ftplib import FTP, error_perm, error_temp
from webdav3.client import Client as WebDavClient
from pod2gen import Podcast, Episode, Media, Category, Person
import unicodedata

def load_config():
    dir_path = os.path.dirname(os.path.abspath(__file__))  # Pobiera ścieżkę katalogu, w którym znajduje się skrypt
    config_path = os.path.join(dir_path, 'config.json')    # Łączy tę ścieżkę z nazwą pliku config.json
    with open(config_path, 'r') as file:
        return json.load(file)

def load_stations():
    dir_path = os.path.dirname(os.path.abspath(__file__))  # Pobiera ścieżkę katalogu, w którym znajduje się skrypt
    stations_path = os.path.join(dir_path, 'stations.json')    # Łączy tę ścieżkę z nazwą pliku stations.json
    with open(stations_path, 'r') as file:
        return json.load(file)

def normalize_filename(name):
    # Zamiana polskich znaków na ich łacińskie odpowiedniki
    mapping = str.maketrans('ąćęłńóśźżĄĆĘŁŃÓŚŹŻ', 'acelnoszzACELNOSZZ')
    normalized_name = name.translate(mapping)
    # Usuwanie niedozwolonych znaków, ale pozostawienie spacji
    normalized_name = ''.join(c for c in normalized_name if c.isalnum() or c in [' ', '-', '.'])
    return normalized_name

def manage_local_space(directory, config):
    local_max_files = config['local_audio_config']['local_max_files']

    # Jeśli max_files jest równy 0, nie usuwaj żadnych plików
    if local_max_files == 0:
        return

    # Pobierz listę plików w katalogu
    files = [os.path.join(directory, f) for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    audio_files = [f for f in files if f.endswith('.mp3') or f.endswith('.opus') or f.endswith('.m4a')]

    # Jeśli liczba plików przekracza max_files, usuń najstarsze
    while len(audio_files) > local_max_files:
        oldest_file = min(audio_files, key=os.path.getmtime)
        os.remove(oldest_file)
        print(f"Usunięto najstarszy plik lokalny: {oldest_file}")
        audio_files.remove(oldest_file)

def manage_ftp_space(ftp, folder_path, config):
    ftp_max_files = config['ftp_config']['ftp_max_files']

    # Jeśli max_files jest równy 0, nie usuwaj żadnych plików
    if ftp_max_files == 0:
        return

    ftp.cwd(folder_path)
    files = ftp.nlst()
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    while len(audio_files) > ftp_max_files:
        oldest_file = min(audio_files, key=lambda x: ftp.sendcmd('MDTM ' + x))
        ftp.delete(oldest_file)
        print(f"Usunięto najstarszy plik: {oldest_file}")
        audio_files.remove(oldest_file)

def parse_date_from_filename(filename):
    # Zakładamy, że nazwa pliku ma format: "ShowName - StationName (DD.MM.YY HH.MM)"
    date_str = filename.split('(')[-1].split(')')[0]  # Pobieramy część pomiędzy nawiasami
    return datetime.strptime(date_str, '%d.%m.%y %H.%M')

def rss(ftp_config, remote_dir, station, show_name):
    ftp_host = ftp_config['ftp_hostname']
    ftp_username = ftp_config['ftp_username']
    ftp_password = ftp_config['ftp_password']
    web_base_url = ftp_config['web_base_url']

    podcast = Podcast()
    podcast.name = f"{show_name} - {station}"
    podcast.authors = [Person(f"{station}")]
    podcast.email = "example@example.com"
    podcast.website = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name)).replace(" ", "%20")
    podcast.description = f"Podcast {show_name} z {station}"
    podcast.language = "pl"
    podcast.category = Category("News", "Politics")
    podcast.explicit = False
    podcast.image = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name), 'podcast.jpg').replace(" ", "%20")

    files_info = []

    with FTP(ftp_host) as ftp:
        ftp.login(ftp_username, ftp_password)
        ftp.cwd(remote_dir)
        ftp.set_pasv(True)
        ftp.sendcmd("TYPE I")
        try:
            items = list(ftp.mlsd())
            for item in items:
                name, info = item
                if info.get('type') == 'file' and (name.endswith('.mp3') or name.endswith('.m4a')):
                    size = info.get('size')
                    files_info.append({'name': name, 'size': size})
        except Exception as e:
            print(f"Nie można uzyskać informacji o plikach: {e}")

    remote_dir_relative = remote_dir.replace(ftp_config['ftp_base_path'], '').lstrip('/')
    web_dir_url = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name))

    for file_info in files_info:
        date = parse_date_from_filename(file_info['name'])
        pub_date = date.strftime('%a, %d %b %Y %H:%M:%S GMT')
        episode = Episode()
        episode_title = f"{show_name} - {station} ({date.strftime('%d.%m.%Y')})"
        episode.title = episode_title
        episode.media = Media(url=f"{web_dir_url.replace(' ', '%20')}/{file_info['name'].replace(' ', '%20')}", size=file_info['size'], type="audio/mpeg")
        episode.publication_date = pub_date
        episode.summary = 'Opis odcinka'
        podcast.episodes.append(episode)

    rss_content = podcast.rss_str()

    local_rss_file = 'feed.xml'
    with open(local_rss_file, 'wb') as f:
        f.write(rss_content.encode('utf-8'))

    with FTP(ftp_host) as ftp:
        ftp.login(ftp_username, ftp_password)
        ftp.cwd(remote_dir)
        with open(local_rss_file, 'rb') as f:
            ftp.storbinary(f'STOR {os.path.basename(local_rss_file)}', f)
    os.remove(local_rss_file)

    print(f"Wygenerowano i przesłano plik RSS: {local_rss_file} do katalogu: {remote_dir}")

def manage_nextcloud_space(client, folder_path, config, station, show_name):
    nextcloud_max_files = config['nextcloud_config']['nextcloud_max_files']
    if nextcloud_max_files == 0:
        return

    full_folder_path = folder_path
    files = client.list(full_folder_path)
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]
    full_files_paths = [os.path.join(full_folder_path, file) for file in audio_files]
    files_info = [(file, client.info(file)) for file in full_files_paths]

    if len(files_info) > nextcloud_max_files:
        files_sorted = sorted(files_info, key=lambda x: x[1].get('modified'))
        for file_path, _ in files_sorted[:-nextcloud_max_files]:
            client.clean(file_path)
            print(f"Usunięto najstarszy plik w Nextcloud: {file_path}")

def upload_ftp(filename, config, station, show_name):
    attempts = 0
    success = False
    while not success and attempts < 5:
        try:
            with FTP(config['ftp_config']['ftp_hostname']) as ftp:
                ftp.login(config['ftp_config']['ftp_username'], config['ftp_config']['ftp_password'])
                remote_dir = f"{config['ftp_config']['ftp_base_path']}/{normalize_filename(station)}/{normalize_filename(show_name)}"

                parts = remote_dir.split('/')
                current_dir = ''
                for part in parts:
                    if part:
                        current_dir += f'/{part}'
                        try:
                            ftp.cwd(current_dir)
                        except error_perm as e:
                            if 'No such file or directory' in str(e):
                                ftp.mkd(current_dir)
                                ftp.cwd(current_dir)
                            else:
                                raise

                with open(filename, 'rb') as file:
                    ftp.storbinary('STOR ' + os.path.basename(filename), file)
                success = True
                manage_ftp_space(ftp, remote_dir, config)
                rss(config['ftp_config'], remote_dir, station, show_name)
                print(f"Plik {filename} został pomyślnie przesłany na FTP.")
        except (error_temp, BrokenPipeError) as e:
            attempts += 1
            print(f"Wystąpił błąd: {e}. Próba ponownego połączenia i przesłania... ({attempts})")
        except Exception as e:
            print(f"Wystąpił nieoczekiwany błąd: {e}")
            break

    if not success:
        print("Nie udało się przesłać pliku po kilku próbach.")

def ensure_remote_path(client, remote_path):
    parts = remote_path.strip('/').split('/')
    path = ''
    for part in parts[:-1]:
        path += f"/{part}"
        if not client.check(path):
            client.mkdir(path)

def upload_nextcloud(filename, config, station, show_name):
    options = {
        'webdav_hostname': config['nextcloud_config']['nextcloud_hostname'],
        'webdav_login': config['nextcloud_config']['nextcloud_login'],
        'webdav_password': config['nextcloud_config']['nextcloud_password']
    }
    client = WebDavClient(options)
    remote_dir = f"/public/{normalize_filename(station)}/{normalize_filename(show_name)}/"
    remote_path = remote_dir + os.path.basename(filename)

    ensure_remote_path(client, remote_path)
    client.upload_sync(remote_path=remote_path, local_path=filename)
    manage_nextcloud_space(client, remote_dir, config, station, show_name)
    print(f"Plik {filename} został pomyślnie przesłany na NextCloud.")

def record_show(duration, station_url, station, show_name, config):
    recordings_path = config['local_audio_config'].get('local_recordings_path', './')
    if recordings_path == './':
        dir_path = os.path.dirname(os.path.abspath(__file__))
    else:
        # Zmiana tutaj: interpretacja "./" jako względna do lokalizacji skryptu, nie katalogu roboczego
        dir_path = os.path.abspath(recordings_path) if not recordings_path.startswith('./') else os.path.join(os.path.dirname(os.path.abspath(__file__)), recordings_path[2:])


    local_audio_config = config['local_audio_config']
    date_str = datetime.now().strftime("%d.%m.%y %H.%M")

    output_dir = os.path.join(dir_path, "recordings", normalize_filename(station), normalize_filename(show_name))
    os.makedirs(output_dir, exist_ok=True)

    # Nowy format nazwy pliku
    output_filename = os.path.join(output_dir, f"{normalize_filename(show_name)} - {normalize_filename(station)} ({date_str}).{local_audio_config['local_audio_format']}")
    print(f"Output filename: {output_filename}")

    if local_audio_config['local_audio_format'] == 'opus':
        codec = "libopus"
    elif local_audio_config['local_audio_format'] == 'mp3':
        codec = "libmp3lame"
    elif local_audio_config['local_audio_format'] == 'm4a':
        codec = "aac"
    else:
        codec = "copy"

    ffmpeg_command = [
        "ffmpeg",
        "-t", str(duration * 60),
        "-i", station_url,
        "-acodec", codec,
        "-b:a", f"{local_audio_config['local_audio_bitrate']}k",
        "-ac", str(local_audio_config['local_audio_channels']),
        "-y", output_filename
    ]
    subprocess.run(ffmpeg_command)
    return output_filename


def convert_audio(source_file, config, destination):
    target_config = config['ftp_config'] if destination == 'ftp' else config['nextcloud_config']

    target_format = target_config['ftp_audio_format'] if destination == 'ftp' else target_config['nextcloud_audio_format']
    target_bitrate = target_config['ftp_audio_bitrate'] if destination == 'ftp' else target_config['nextcloud_audio_bitrate']
    channels = target_config['ftp_audio_channels'] if destination == 'ftp' else target_config['nextcloud_audio_channels']

    target_filename = os.path.splitext(source_file)[0] + f' {destination}.{target_format}'
    print(f"Target filename: {target_filename}")

    if not (config['local_audio_config']['local_audio_format'] == target_format and
            config['local_audio_config']['local_audio_bitrate'] == target_bitrate and
            config['local_audio_config']['local_audio_channels'] == channels):
        print(f"Konwertowanie pliku do formatu {target_format} z bitrate: {target_bitrate}k i {channels} kanałów.")

        if target_format == 'opus':
            codec = "libopus"
            format_output = "opus"
        elif target_format == 'mp3':
            codec = "libmp3lame"
            format_output = "mp3"
        elif target_format == 'm4a':
            codec = "aac"
            format_output = "ipod"
        else:
            codec = "copy"
            format_output = target_format

        ffmpeg_command = [
            "ffmpeg",
            "-i", source_file,
            "-acodec", codec,
            "-b:a", f"{target_bitrate}k",
            "-ac", str(channels),
            "-f", format_output,
            "-y", target_filename
        ]
        subprocess.run(ffmpeg_command, check=True)
    else:
        print("Plik wejściowy jest już w docelowym formacie i jakości. Tworzenie kopii.")
        shutil.copy(source_file, target_filename)

    return target_filename

def convert_audio(source_file, config, destination):
    target_config = config['ftp_config'] if destination == 'ftp' else config['nextcloud_config']

    target_format = target_config['ftp_audio_format'] if destination == 'ftp' else target_config['nextcloud_audio_format']
    target_bitrate = target_config['ftp_audio_bitrate'] if destination == 'ftp' else target_config['nextcloud_audio_bitrate']
    channels = target_config['ftp_audio_channels'] if destination == 'ftp' else target_config['nextcloud_audio_channels']

    target_filename = os.path.splitext(source_file)[0] + f' {destination}.{target_format}'
    print(f"Target filename: {target_filename}")

    if not (config['local_audio_config']['local_audio_format'] == target_format and
            config['local_audio_config']['local_audio_bitrate'] == target_bitrate and
            config['local_audio_config']['local_audio_channels'] == channels):
        print(f"Konwertowanie pliku do formatu {target_format} z bitrate: {target_bitrate}k i {channels} kanałów.")

        if target_format == 'opus':
            codec = "libopus"
            format_output = "opus"
        elif target_format == 'mp3':
            codec = "libmp3lame"
            format_output = "mp3"
        elif target_format == 'm4a':
            codec = "aac"
            format_output = "ipod"
        else:
            codec = "copy"
            format_output = target_format

        ffmpeg_command = [
            "ffmpeg",
            "-i", source_file,
            "-acodec", codec,
            "-b:a", f"{target_bitrate}k",
            "-ac", str(channels),
            "-f", format_output,
            "-y", target_filename
        ]
        subprocess.run(ffmpeg_command, check=True)
    else:
        print("Plik wejściowy jest już w docelowym formacie i jakości. Tworzenie kopii.")
        shutil.copy(source_file, target_filename)

    return target_filename



def main():
    parser = ArgumentParser()
    parser.add_argument("station", type=str, help="Station name")
    parser.add_argument("show_name", type=str, help="Name of the show")
    parser.add_argument("duration", type=int, help="Duration of the recording in minutes")
    parser.add_argument("-s", "--save", nargs='+', choices=['ftp', 'nextcloud', 'local'], help="Save destinations", default=[])
    args = parser.parse_args()

    config = load_config()
    stations = load_stations()
    station_url = stations[args.station]

    original_file = record_show(args.duration, station_url, args.station, args.show_name, config)

    if 'ftp' in args.save:
        ftp_file = convert_audio(original_file, config, 'ftp')
        upload_ftp(ftp_file, config, args.station, args.show_name)
        os.remove(ftp_file)

    if 'nextcloud' in args.save:
        nextcloud_file = convert_audio(original_file, config, 'nextcloud')
        upload_nextcloud(nextcloud_file, config, args.station, args.show_name)
        if nextcloud_file != original_file:
            os.remove(nextcloud_file)

    if 'local' in args.save:
        output_dir = os.path.dirname(original_file)
        manage_local_space(output_dir, config)

    if 'local' not in args.save:
        os.remove(original_file)

if __name__ == "__main__":
    main()
